def chatml(messages):
    content = ""
    for message in messages:
        if message["role"] == "system":
            content += "<|im_start|>system\n" + message["content"] + "<|im_end|>\n"
        elif message["role"] == "user":
            content += "<|im_start|>user\n" + message["content"] + "<|im_end|>\n"
        elif message["role"] == "assistant":
            content += "<|im_start|>assistant\n" + message["content"] + "<|im_end|>\n"
    content += "<|im_start|>assistant\n"
    return content



import threading
from typing import List, Dict, Any, Callable
from tqdm import tqdm  # 导入 tqdm
from .base import ParallelBase

class MultiThread(ParallelBase):
    def __init__(self, parallel_count: int = 1):
        """
        初始化 MultiThread 类，设定最大并发线程数。
        :param parallel_count: 最大并发线程数，默认为1
        """
        super().__init__(parallel_count)

    def __call__(self, func: Callable[..., Any]) -> Callable[..., List[Any]]:
        """
        使得 MultiThread 实例可以作为装饰器使用。
        :param func: 被装饰的函数
        :return: 一个新的函数，具有并发执行功能
        """
        def wrapper(parallel_params: List[Dict[Any, Any]], *args, **kwargs) -> List[Any]:
            result = [None] * len(parallel_params)
            threads = []

            # 定义每个线程执行的工作
            def worker(index, param):
                result[index] = func(**param)

            # 使用 tqdm 创建进度条，func.__name__ 获取函数名称
            with tqdm(total=len(parallel_params), desc=f"Multi Thread Processing -> [{func.__name__}]") as pbar:
                # 创建并启动线程，限制并行线程数
                for i, param in enumerate(parallel_params):
                    if len(threads) >= self.parallel_count:
                        threads[0].join()  # 等待第一个线程完成后再启动新线程
                        threads.pop(0)

                    thread = threading.Thread(target=worker, args=(i, param))
                    threads.append(thread)
                    thread.start()

                    # 每启动一个线程，就更新进度条
                    pbar.update(1)

                # 等待所有线程完成
                for thread in threads:
                    thread.join()

            return result

        return wrapper



import time
import inspect
import threading
from typing import Callable, Any

from llmer.runtime.exceptions import AcquireLockTimeoutError, ExecutionTimeoutError



def timeout(seconds: float):
    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        def wrapper(*args, timeout=None, **kwargs) -> Any:
            """包装函数，支持动态超时时间"""
            effective_timeout = timeout if timeout is not None else seconds
            is_generator = inspect.isgeneratorfunction(func)

            if is_generator:
                def generator_wrapper():
                    start_time = time.time()  # 记录第一次 yield 开始时间
                    thread_exception = None

                    # 启动线程来执行生成器
                    def worker():
                        nonlocal thread_exception
                        try:
                            for item in func(*args, **kwargs):
                                yield item
                        except Exception as e:
                            thread_exception = e
                            raise e

                    # 生成器实例
                    gen = worker()
                    while True:
                        try:
                            # 每次调用生成器时都要检查超时
                            elapsed_time = time.time() - start_time
                            if elapsed_time > effective_timeout:
                                raise ExecutionTimeoutError(
                                    f"Function '{func.__name__}' timed out after {effective_timeout} seconds"
                                )

                            # `next(gen)` 会迭代到下一个生成器项
                            yield next(gen)
                        except StopIteration:
                            break  # 如果生成器完成，退出循环
                        except Exception as e:
                            thread_exception = e
                            break

                    if thread_exception:
                        raise thread_exception

                return generator_wrapper()
            else:
                result = []
                thread_exception = None

                def worker():
                    nonlocal thread_exception
                    try:
                        result.append(func(*args, **kwargs))
                    except Exception as e:
                        thread_exception = e

                thread = threading.Thread(target=worker, daemon=True)
                thread.start()
                thread.join(timeout=effective_timeout)

                if thread.is_alive():
                    raise ExecutionTimeoutError(
                        f"Function '{func.__name__}' timed out after {effective_timeout} seconds"
                    )

                if thread_exception:
                    raise thread_exception

                return result[0] if result else None

        return wrapper
    return decorator




def parallel_safe_lock(lock: threading.Lock, seconds: float = 1) -> Callable[..., Any]:

    def decorator(rw_func: Callable[..., Any]) -> Callable[..., Any]:

        def wrapper(*args, timeout=None, **kwargs) -> Any:
            """
            包装函数：允许在每次调用时覆盖超时值
            :param timeout: 覆盖超时时间
            :param args: 被装饰函数的参数
            :param kwargs: 被装饰函数的关键字参数
            :return: 函数的返回结果
            """
            # 如果调用时有传递覆盖的超时值，则使用覆盖值
            lock_timeout = timeout if timeout is not None else seconds
            acquired = False
            try:
                if not lock.acquire(timeout=lock_timeout):
                    raise AcquireLockTimeoutError(f"{rw_func.__name__} acquires lock, timeout exceeded {lock_timeout}")
                acquired = True
                return rw_func(*args, **kwargs)
            finally:
                if acquired:
                    lock.release()

        return wrapper

    return decorator






from .multi_thread import MultiThread
from .thread_pool import ThreadPool
from .async_parallel import AsyncParallel
from .async_executor import AsyncExecutor



__all__ = [
    'AsyncParallel',
    'AsyncExecutor',
    'MultiThread',
    'ThreadPool'
]


import os
import sys
import base64
import yaml
import json
import jsonlines
from typing import List, Any, Dict, Optional

def get_main_script_dir():
    """获取最终执行脚本的目录"""
    # sys.argv[0] 包含了运行时执行的脚本路径
    main_script = sys.argv[0]
    main_script_dir = os.path.dirname(os.path.abspath(main_script))
    return main_script_dir



def file_to_list(path: Optional[str] = None):
    res = []
    # 获取最终执行脚本的目录
    main_script_dir = get_main_script_dir()

    # 如果传入的路径不是绝对路径，则与最终执行脚本的目录拼接
    if path and not os.path.isabs(path):
        path = os.path.join(main_script_dir, path)

    with jsonlines.open(path, "r") as r:
        for line in r:
            res.append(line)
    return res



def list_to_file(data: List[Any] = None, path: Optional[str] = None, mode='w'):
    """将列表保存为 JSONL 文件"""
    # 获取最终执行脚本的目录
    main_script_dir = get_main_script_dir()

    # 如果传入的路径不是绝对路径，则与最终执行脚本的目录拼接
    if path and not os.path.isabs(path):
        path = os.path.join(main_script_dir, path)
    if data:
        with jsonlines.open(path, mode=mode) as writer:
            writer.write_all(data)


def yaml_reader(path: Optional[str] = None) -> Dict[str, Any]:
    """读取 YAML 配置文件并返回内容"""
    # 获取最终执行脚本的目录
    main_script_dir = get_main_script_dir()

    # 如果传入的路径是相对路径，则与脚本目录拼接成绝对路径
    if not os.path.isabs(path):
        path = os.path.join(main_script_dir, 'prompts', path)

    # 使用安全加载方法读取 YAML 文件
    with open(path, 'r') as file:
        component = yaml.safe_load(file)

    return component



def image_to_base64(path: Optional[str] = None, prefix=False) -> str:
    """
    将图片文件转换为 Base64 编码的字符串，并添加适当的前缀。

    :param path: 图片的路径（相对或绝对路径）。
    :return: 带有前缀的 Base64 编码字符串。
    """
    # 获取最终执行脚本的目录
    main_script_dir = get_main_script_dir()

    # 如果传入的路径不是绝对路径，则与最终执行脚本的目录拼接
    if path and not os.path.isabs(path):
        path = os.path.join(main_script_dir, path)

    # 获取图片的扩展名
    _, ext = os.path.splitext(path)
    ext = ext.lower().strip('.')  # 获取文件扩展名并转小写

    # 支持的图像格式（根据扩展名来判断）
    valid_image_formats = ["png", "jpeg", "jpg", "gif", "bmp", "tiff", "webp", "ico"]

    if ext not in valid_image_formats:
        raise ValueError(f"Unsupported image format: {ext}. Supported formats are: {', '.join(valid_image_formats)}")

    # 打开图片文件并读取内容
    with open(path, "rb") as img_file:
        # 读取图片并转换为 Base64 编码
        encoded_image = base64.b64encode(img_file.read()).decode("utf-8")

    if prefix:
        # 返回带前缀的 Base64 编码字符串
        data_prefix = f"data:image/{ext};base64,"
        encoded_image = data_prefix + encoded_image


    return encoded_image


from abc import ABC, abstractmethod
from typing import List, Dict, Any, Callable

class ParallelBase(ABC):

    def __init__(self, parallel_count: int = 1):
        self.parallel_count = parallel_count

    def __call__(self, func: Callable[..., Any]) -> Callable[..., List[Any]]:
        raise NotImplementedError("__call__ not implemented!")



from abc import ABC, abstractmethod
from typing import List, Dict, Any, Callable

class ModelBase(ABC):

    def __init__(
            self,
            api_key: str = None,
            endpoint: str = None,
            api_version: str = None,
            headers: Dict[str, Any] = None,
            timeout: float = 600,
            retry: int = 2
    ):
        self.api_key = api_key
        self.endpoint = endpoint
        self.api_version = api_version
        self.headers = headers
        self.timeout = timeout
        self.retry = retry

    def chat(
            self,
            model: str = None,
            temperature: float = 1.0,
            top_p: float = 0.1,
            max_tokens: int = 1024,
            messages: List[Dict[str, Any]] = None,
            response_format = None,
            tools: List = None,
            tool_choice: str = None,
            stream: bool = False
    ):
        if stream:
            return self._stream_response(
                model,
                temperature,
                top_p,
                max_tokens,
                messages,
                response_format,
                tools,
                tool_choice
            )
        else:
            return self._non_stream_response(
                model,
                temperature,
                top_p,
                max_tokens,
                messages,
                response_format,
                tools,
                tool_choice
            )

    def _non_stream_response(self, *args, **kwargs):
        raise NotImplementedError("_non_stream_response not implemented!")

    def _stream_response(self, *args, **kwargs):
        raise NotImplementedError("_stream_response not implemented!")



import asyncio
from typing import Callable, List, Dict, Any
from tqdm.asyncio import tqdm  # 导入 tqdm.asyncio 来支持异步更新
from .base import ParallelBase

class AsyncParallel(ParallelBase):
    def __init__(self, parallel_count: int = 1):
        """
        初始化 AsyncParallel 类，设定最大并发任务数
        :param parallel_count: 最大并发任务数，默认为1
        """
        super().__init__(parallel_count)
        self.semaphore = asyncio.Semaphore(parallel_count)  # 控制并发数量

    def __call__(self, func: Callable[..., Any]) -> Callable[..., List[Any]]:
        """
        使 AsyncParallel 实例能够作为装饰器使用。
        :param func: 被装饰的函数
        :return: 一个新的异步函数，支持并发执行
        """
        async def wrapper(parallel_params: List[Dict[Any, Any]], *args, **kwargs) -> List[Any]:
            result = [None] * len(parallel_params)

            # 创建 tqdm 进度条
            with tqdm(total=len(parallel_params), desc=f"Async Parallel Processing -> [{func.__name__}]") as pbar:
                async def worker(index, param):
                    async with self.semaphore:  # 限制并发任务数
                        result[index] = await func(**param)  # 调用被装饰的函数
                        pbar.update(1)  # 每当一个任务完成时，更新进度条

                # 使用 asyncio.gather 并发执行所有任务
                await asyncio.gather(*[
                    worker(i, param) for i, param in enumerate(parallel_params)
                ])

            return result

        return wrapper



import asyncio
from typing import Callable, Any, List, Dict
from tqdm.asyncio import tqdm  # 导入 tqdm.asyncio 来支持异步更新
from .base import ParallelBase

class AsyncExecutor(ParallelBase):
    def __init__(self, parallel_count: int = 1):
        """
        初始化 AsyncExecutor 类，设定最大并发数量
        :param parallel_count: 最大并发任务数，默认为1
        """
        super().__init__(parallel_count)
        self.semaphore = asyncio.Semaphore(parallel_count)  # 限制最大并发数

    def __call__(self, func: Callable[..., Any]) -> Callable[..., List[Any]]:
        """
        使得 AsyncExecutor 实例可以作为装饰器使用。
        :param func: 被装饰的函数
        :return: 一个新的异步函数，支持并发执行
        """
        async def wrapper(parallel_params: List[Dict[Any, Any]], *args, **kwargs) -> List[Any]:
            result = [None] * len(parallel_params)

            # 使用 tqdm 创建进度条
            with tqdm(total=len(parallel_params), desc=f"Async Executor Processing -> [{func.__name__}]") as pbar:
                async def worker(index, param):
                    async with self.semaphore:  # 限制并发
                        loop = asyncio.get_running_loop()  # 获取当前事件循环
                        # 使用 run_in_executor 将非异步函数放入线程池
                        result[index] = await loop.run_in_executor(
                            None,  # 使用默认的 ThreadPoolExecutor
                            func,   # 要调用的非异步函数
                            *param.values()  # 解包参数字典
                        )
                        pbar.update(1)  # 每完成一个任务，更新进度条

                # 使用 asyncio.gather 并发执行所有任务
                await asyncio.gather(*[
                    worker(i, param) for i, param in enumerate(parallel_params)
                ])

            return result

        return wrapper



import concurrent.futures
from typing import List, Dict, Any, Callable
from tqdm import tqdm  # 导入 tqdm 库
from .base import ParallelBase

class ThreadPool(ParallelBase):
    def __init__(self, parallel_count: int = 1):
        """
        初始化 ThreadPool 类，设定最大并发线程数。
        :param parallel_count: 最大并发线程数，默认为1
        """
        super().__init__(parallel_count)

    def __call__(self, func: Callable[..., Any]) -> Callable[..., List[Any]]:
        """
        使得 ThreadPool 实例可以作为装饰器使用。
        :param func: 被装饰的函数
        :return: 一个新的函数，具有并发执行功能
        """
        def wrapper(parallel_params: List[Dict[Any, Any]], *args, **kwargs) -> List[Any]:
            result = [None] * len(parallel_params)

            # 使用 ThreadPoolExecutor 来并行执行任务
            with concurrent.futures.ThreadPoolExecutor(max_workers=self.parallel_count) as executor:
                # 提交任务并记录对应的索引
                futures = {
                    executor.submit(func, **param): index
                    for index, param in enumerate(parallel_params)
                }

                # 使用 tqdm 显示进度条
                for future in tqdm(concurrent.futures.as_completed(futures), total=len(futures), desc=f"Thread Pool Processing -> [{func.__name__}]"):
                    index = futures[future]
                    result[index] = future.result()

            return result

        return wrapper



import openai
from typing import List, Dict, Any, Callable

from .base import ModelBase

class Azure(ModelBase):
    def __init__(
            self,
            api_key: str = None,
            endpoint: str = None,
            api_version: str = None,
            headers: Dict[str, Any] = None,
            timeout: float = 600,
            retry: int = 2
    ):
        super().__init__(
            api_key=api_key,
            endpoint=endpoint,
            api_version=api_version,
            headers=headers,
            timeout=timeout,
            retry=retry
        )

        self.client = openai.AzureOpenAI(
            default_headers=self.headers,
            api_key = self.api_key,
            api_version = self.api_version,
            azure_endpoint = self.endpoint,
            timeout = self.timeout,
            max_retries = self.retry
        )

    def _non_stream_response(
            self,
            model,
            temperature,
            top_p,
            max_tokens,
            messages,
            response_format,
            tools,
            tool_choice
    ):
        response = self.client.chat.completions.create(
            model=model,
            temperature=temperature,
            top_p=top_p,
            max_tokens=max_tokens,
            messages=messages,
            response_format=response_format,
            tools=tools,
            tool_choice=tool_choice,
            stream=False
        )
        return response

    def _stream_response(
            self,
            model,
            temperature,
            top_p,
            max_tokens,
            messages,
            response_format,
            tools,
            tool_choice
    ):
        response = self.client.chat.completions.create(
            model=model,
            temperature=temperature,
            top_p=top_p,
            max_tokens=max_tokens,
            messages=messages,
            response_format=response_format,
            tools=tools,
            tool_choice=tool_choice,
            stream=True
        )
        for chunk in response:
            yield chunk






from setuptools import setup, find_packages
import pathlib

# 读取 README 文件内容
here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README_ZH.MD').read_text(encoding='utf-8')

# 读取版本号
about = {}
exec((here / 'llmer' / '__version__.py').read_text(), about)

setup(
    name='llmer',
    version=about['__version__'],  # 动态获取版本号
    packages=find_packages(),
    description='llmer is a lightweight Python library designed to streamline the development of applications leveraging large language models (LLMs). It provides high-level APIs and utilities for parallel processing, runtime management, file handling, and prompt generation, reducing the overhead of repetitive tasks.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='pydaxing',
    author_email='pydaxing@gmail.com',
    url='https://github.com/pydaxing/llmer',
    python_requires='>=3.6',
    install_requires=[
        'jsonlines',
        'PyYAML',
        'setuptools',
        'tqdm',
        'openai',
    ],
    entry_points={},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Operating System :: OS Independent',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Typing :: Typed',
    ],
    keywords='llmer leveraging large language models',
)



class LLMExectionError(Exception):
    error_code = 95101
    error_msg = 'LLM Execution Failed!'

    def __init__(self, message=None, *args):
        super().__init__(message, *args)
        if message:
            self.error_msg += f' {message}'

    def __str__(self):
        return self.error_msg



class AgentExecutionError(Exception):
    error_code = 95102
    error_msg = 'Agent Execution Failed!'

    def __init__(self, message=None, *args):
        super().__init__(message, *args)
        if message:
            self.error_msg += f' {message}'

    def __str__(self):
        return self.error_msg


class ToolExecutionError(Exception):
    error_code = 95103
    error_msg = 'Tool Execution Failed!'

    def __init__(self, message=None, *args):
        super().__init__(message, *args)
        if message:
            self.error_msg += f' {message}'

    def __str__(self):
        return self.error_msg


class ModuleExecutionError(Exception):
    error_code = 95104
    error_msg = 'Module Execution Failed!'

    def __init__(self, message=None, *args):
        super().__init__(message, *args)
        if message:
            self.error_msg += f' {message}'

    def __str__(self):
        return self.error_msg

# 自定义锁超时异常
class AcquireLockTimeoutError(Exception):
    error_code = 95105
    error_msg = 'Lock Acquisition Failed!'

    def __init__(self, message=None, *args):
        super().__init__(message, *args)
        if message:
            self.error_msg += f' {message}'

    def __str__(self):
        return self.error_msg


class ExecutionTimeoutError(Exception):
    """自定义异常，指示函数执行超时"""
    error_code = 95106
    error_msg = 'Execution Time Out!'

    def __init__(self, message=None, *args):
        super().__init__(message, *args)
        if message:
            self.error_msg += f' {message}'

    def __str__(self):
        return self.error_msg




import time
import asyncio

from llmer.parallel.async_parallel import AsyncParallel
from llmer.parallel.thread_pool import ThreadPool
from llmer.parallel.multi_thread import MultiThread
from llmer.parallel.async_executor import AsyncExecutor


@ThreadPool(parallel_count=4)
def thread_pool_sample_function(x, y):
    time.sleep(1.5)
    return x*y['z']**2

@MultiThread(parallel_count=4)
def multi_thread_sample_function(x, y):
    time.sleep(1.5)
    return x*y['z']**2

@AsyncExecutor(parallel_count=4)
def async_executor_sample_function(x, y):
    time.sleep(1.5)
    return x*y['z']**2

@AsyncParallel(parallel_count=4)
async def async_parallel_sample_function(x: int, y: int) -> int:
    await asyncio.sleep(1.5)
    return x*y['z']**2


class TestLLMUParallel:
    params = [{'x': 1, 'y': {'z': 2} }, {'x': 2, 'y': {'z': 3}}, {'x': 3, 'y': {'z': 4}}, {'x': 4, 'y': {'z': 5}}, {'x': 5, 'y': {'z': 6}}, {'x': 6, 'y': {'z': 7}}]

    def test(self):
        thread_pool_start = time.time()
        results = thread_pool_sample_function(parallel_params=self.params)
        thread_pool_spend = time.time() - thread_pool_start
        assert 3 <= thread_pool_spend < 3.5, f"ThreadPool time spent: {thread_pool_spend} seconds"
        print("ThreadPool Test Successfully")
        time.sleep(1)

        multi_thread_start = time.time()
        results = multi_thread_sample_function(self.params)
        multi_thread_spend = time.time() - multi_thread_start
        assert 3 <= multi_thread_spend < 3.5, f"MultiThread time spent: {multi_thread_spend} seconds."
        print("MultiThread Test Successfully")
        time.sleep(1)


    async def async_test(self):
        async_executor_start = time.time()
        results = await async_executor_sample_function(self.params)
        async_executor_spend = time.time() - async_executor_start
        assert 3 <= async_executor_spend < 3.5, f"AsyncExecutor time spent: {async_executor_spend} seconds."
        print("AsyncExecutor Test Successfully")
        await asyncio.sleep(1)

        async_parallel_start = time.time()
        results = await async_parallel_sample_function(self.params)
        async_parallel_spend = time.time() - async_parallel_start
        assert 3 <= async_parallel_spend < 3.5, f"AsyncParallel time spent: {async_parallel_spend} seconds."
        print("AsyncParallel Test Successfully")
        await asyncio.sleep(1)



    def run_tests(self):
        self.test()
        asyncio.run(self.async_test())

TestLLMUParallel().run_tests()



